package Array;

use strict;
use Carp;
use Scalar::Util 'blessed';

sub new {
	my $class = shift;
	my %args = @_;
	my $self = {};

	foreach (keys %args) {
		$self->{$_} = $args{$_};
		#print __PACKAGE__ . ": " . $_ . ": " . $args{$_} . "\n";
	}
	
	# array with nested struct|array inside
	if ($args{nested} && blessed($args{nested})) {
		# count
		if($args{count}) {
			# count is a number (fixed)
			if($args{count} =~ /^\d+$/) {
				# name
				if($args{name}) {
					$self->{pack} = sub : method {
						my ($self, $chash, $bin) = @_;
						foreach (@{$chash->{$self->{name}}}) {
							$self->{nested}->pack($_, $bin);
						}
					};
					if ($args{nested}->isa("Struct")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $self->{count}; $i++) {
								$self->{nested}->unpack(\%{$chash->{$self->{name}}[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Array")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $self->{count}; $i++) {
								$self->{nested}->unpack(\@{$chash->{$self->{name}}[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Scalar")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $self->{count}; $i++) {
								$self->{nested}->unpack(\$chash->{$self->{name}}[$i], $bin, $position);
							}
						}
					}
				# no name
				} else {
					$self->{pack} = sub : method {
						my ($self, $chash, $bin) = @_;
						foreach (@{$chash}) {
							$self->{nested}->pack($_, $bin);
						}
					};
					if ($args{nested}->isa("Struct")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $self->{count}; $i++) {
								$self->{nested}->unpack(\%{$chash->[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Array")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $self->{count}; $i++) {
								$self->{nested}->unpack(\@{$chash->[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Scalar")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $self->{count}; $i++) {
								$self->{nested}->unpack(\$chash->[$i], $bin, $position);
							}
						}
					}
				}
			# count is text (dynamic, found in nest parent)
			} elsif ($args{count}  =~ /^\w+$/) {
				# name
				if($args{name}) {
					$self->{pack} = sub : method {
						my ($self, $chash, $bin) = @_;
						foreach (@{$chash->{$self->{name}}}) {
							$self->{nested}->pack($_, $bin);
						}
					};
					if($args{nested}->isa("Struct")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $chash->{$self->{count}}; $i++) {
								$self->{nested}->unpack(\%{$chash->{$self->{name}}[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Array")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $chash->{$self->{count}}; $i++) {
								$self->{nested}->unpack(\@{$chash->{$self->{name}}[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Scalar")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $chash->{$self->{count}}; $i++) {
								$self->{nested}->unpack(\$chash->{$self->{name}}[$i], $bin, $position);
							}
						}
					}
				# no name
				} else {
					$self->{pack} = sub : method {
						my ($self, $chash, $bin) = @_;
						foreach (@{$chash}) {
							$self->{nested}->pack($_, $bin);
						}
					};
					if ($args{nested}->isa("Struct")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $chash->{$self->{count}}; $i++) {
								$self->{nested}->unpack(\%{$chash->[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Array")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $chash->{$self->{count}}; $i++) {
								$self->{nested}->unpack(\@{$chash->[$i]}, $bin, $position);
							}
						}
					} elsif($args{nested}->isa("Scalar")) {
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
							for(my $i = 0; $i < $chash->{$self->{count}}; $i++) {
								$self->{nested}->unpack(\$chash->[$i], $bin, $position);
							}
						}
					}
				}
			}
		# slurp
		} else {
			# name
			if($args{name}) {
				$self->{pack} = sub : method {
					my ($self, $chash, $bin) = @_;
					foreach (@{$chash->{$self->{name}}}) {
						$self->{nested}->pack($_, $bin);
					}
				};
				if($args{nested}->isa("Struct")) {
					$self->{unpack} = sub : method {
						my ($self, $chash, $bin, $position) = @_;
						my $len = length($$bin);
						for (my $i = 0; $$position < $len; $i++) {
							$self->{nested}->unpack(\%{$chash->{$self->{name}}[$i]}, $bin, $position);
						}
					}
				} elsif($args{nested}->isa("Array")) {
					$self->{unpack} = sub : method {
						my ($self, $chash, $bin, $position) = @_;
						my $len = length($$bin);
						for (my $i = 0; $$position < $len; $i++) {
							$self->{nested}->unpack(\@{$chash->{$self->{name}}[$i]}, $bin, $position);
						}
					}
				} elsif($args{nested}->isa("Scalar")) {
					$self->{unpack} = sub : method {
						my ($self, $chash, $bin, $position) = @_;
						my $len = length($$bin);
						for (my $i = 0; $$position < $len; $i++) {
							$self->{nested}->unpack(\{$chash->{$self->{name}}[$i]}, $bin, $position);
						}
					}
				}
			# no name
			} else {
				$self->{pack} = sub : method {
					my ($self, $chash, $bin) = @_;
					foreach (@{$chash}) {
						$self->{nested}->pack($_, $bin);
					}
				};
				if ($args{nested}->isa("Struct")) {
					$self->{unpack} = sub : method {
						my ($self, $chash, $bin, $position) = @_;
						my $len = length($$bin);
						for (my $i = 0; $$position < $len; $i++) {
							$self->{nested}->unpack(\%{$chash->[$i]}, $bin, $position);
						}
					}
				} elsif($args{nested}->isa("Array")) {
					$self->{unpack} = sub : method {
						my ($self, $chash, $bin, $position) = @_;
						my $len = length($$bin);
						for (my $i = 0; $$position < $len; $i++) {
							$self->{nested}->unpack(\@{$chash->[$i]}, $bin, $position);
						}
					}
				} elsif($args{nested}->isa("Scalar")) {
					$self->{unpack} = sub : method {
						my ($self, $chash, $bin, $position) = @_;
						my $len = length($$bin);
						for (my $i = 0; $$position < $len; $i++) {
							$self->{nested}->unpack(\{$chash->[$i]}, $bin, $position);
						}
					}
				}
			}
		}
	} else {
		# valid simple array
		# TODO: we don't need both name and key, as they have the same function
		# TODO: do we actually use this?
		if ($args{type} && $args{len}) {
			# count
			if($args{count}) {
				# we do the static calculations here, so we won't have to do them at runtime
				$self->{type} = $args{type} . $args{count}; # 'v' . '4' = v4
				$self->{len} = $args{len} * $args{count}; # 2 * 4 = 8
				delete $self->{count};
				# name
				if($args{name}) {
					# key
					if ($args{key}) {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type}, @{$chash->{$self->{name}}{$self->{key}}});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
				
							@{$chash->{$self->{name}}{$self->{key}}} = unpack($self->{type}, substr($$bin, $$position, $self->{len}));
							$$position += $self->{len};
						
							return 0;
						};
					# no key
					} else {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type}, @{$chash->{$self->{name}}});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
				
							@{$chash->{$self->{name}}} = unpack($self->{type}, substr($$bin, $$position, $self->{len}));
							$$position += $self->{len};
						
							return 0;
						};
					}
				# no name
				} else {
					# key
					if ($args{key}) {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type}, @{$chash->{$self->{key}}});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
				
							@{$chash->{$self->{key}}} = unpack($self->{type}, substr($$bin, $$position, $self->{len}));
							$$position += $self->{len};
						
							return 0;
						};
					# no key
					} else {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type}, @{$chash});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;
				
							@{$chash} = unpack($self->{type}, substr($$bin, $$position, $self->{len}));
							$$position += $self->{len};
						
							return 0;
						};
					}
				}
			# slurp
			} else {
				# name
				if($args{name}) {
					# key
					if ($args{key}) {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type} . @{$chash->{$self->{name}}}, @{$chash->{$self->{name}}{$self->{key}}});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;

							my $len = length($$bin);
							my $count = ($len-$$position)/$self->{len};
						
							@{$chash->{$self->{name}}{$self->{key}}} = unpack($self->{type} . $count, substr($$bin, $$position));
							$$position = $len; # to the end of the packet
						
							return 0;
						};
					# no key
					} else {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type} . @{$chash->{$self->{name}}}, @{$chash->{$self->{name}}});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;

							my $len = length($$bin);
							my $count = ($len-$$position)/$self->{len};
						
							@{$chash->{$self->{name}}} = unpack($self->{type} . $count, substr($$bin, $$position));
							$$position = $len; # to the end of the packet
						
							return 0;
						};
					}
				# no name
				} else {
					# key
					if ($args{key}) {
						$self->{pack} = sub : method {
							my ($self, $chash, $bin) = @_;
			
							$$bin .= pack($self->{type} . @{$chash->{$self->{key}}}, @{$chash->{$self->{key}}});

							return 0;
						};
						$self->{unpack} = sub : method {
							my ($self, $chash, $bin, $position) = @_;

							my $len = length($$bin);
							my $count = ($len-$$position)/$self->{len};
						
							@{$chash->{$self->{key}}} = unpack($self->{type} . $count, substr($$bin, $$position));
							$$position = $len; # to the end of the packet
						
							return 0;
						};
					# no key
					} else {
						croak "This is not possible?";
					}
				}
			}
		# invalid
		} else {
			croak __PACKAGE__ . ": We should not be here! " . ref($args{nested}) . "\n";
		}
	}
	
	return bless $self, $class;
}

sub DESTROY {
	my ($self) = @_;
}

sub pack : method {&{$_[0]->{pack}}(@_);}

sub unpack : method {&{$_[0]->{unpack}}(@_);}

1;