package RO_struct_plugin;

use strict;
use IO::File;
use Data::Dumper;
use Scalar::Util 'blessed';

use Plugins;
use Log qw(message warning error debug);

our $plugin_folder = $Plugins::current_plugin_folder;
use lib $Plugins::current_plugin_folder;

use Struct;
use Array;
use Scalar;

my $structs;
my $struct;
my $i = 0;
my $commentmode = 1;
my $nestchain = 0;
my $neststruct = 0;

my $packetID2HANDLE;

my %type_c2p_lut = (
	'char' => 'c',
	'bool' => 'c', # bool is not a valid ANSI C primitive datatype?
	'short' => 's',
	#'short' => 'v', # TODO: watch out for uc
	'int' => 'i',
	#'int' => 'V',
	'long' => 'l',
	#'long' => 'V',
	'long long' => 'q',
	'float' => 'f',
	'double' => 'd',
	'long double' => 'D',
);

my %len_p2b_lut = (
	'a' => 1,
	'c' => 1,
	'C' => 1,
	'v' => 2,
	's' => 2,
	'S' => 2,
	'V' => 4,
	'i' => 4,
	'I' => 4,
	'l' => 4,
	'L' => 4,
);

my %len_c2b_lut = (
	'char' => 1,
	'bool' => 1,
	'short' => 2,
	'int' => 4,
	'long' => 4,
);

Plugins::register('RO_struct_plugin', 'RO (un)pack test plugin', \&unload, \&reload);
my $hooks = Plugins::addHooks(
	['start3', \&load_data, undef],
	['parseMsg/pre', \&parse_packet, undef],
);

sub unload {
	message "Unloading plugin\n";
	Plugins::delHooks($hooks);
}

sub reload {
	message "Reloading plugin\n";
	$structs = undef;
	&load_data;
}

sub load_data {
	&load_structs;
	&load_headers;
}

sub load_headers {
	my $file = "$plugin_folder/headers.txt";
	my $fh = new IO::File($file, "r") or die "could not open $file: $!\n";
	while (my $line = $fh->getline()) {
		next if ($line =~ /^\/\//); # TODO: improve C comment filtering # we could simply remove /*(.*)*/ but for multiple lines we need to see if there is an unclosed /*
	
		$line =~ s/\x{FEFF}//g;	# Remove BOM
		$line =~ s/^[\t\s]*//;	# Remove leading tabs and whitespace
		$line =~ s/[\r\n]//g;	# Remove line endings
		
		
		#$line =~ s/(__M(\d+))//; # this shouldn't be handled at this layer
		next unless($line =~ /HEADER([\w]*)\s+=\s+0x([0-9A-Fa-f]+)/);
		
		error "packets: " . uc(sprintf("%04s", $2)) . ": ". $1 . "\n"; 
		
		# maybe we can put it in an array
		$packetID2HANDLE->{uc(sprintf("%04s", $2))} = 'PACKET'.$1;
	}
}

sub load_structs {
	my $file = "$plugin_folder/structs.txt";
	my $fh = new IO::File($file, "r") or die "could not open $file: $!\n";
	while (my $line = $fh->getline()) {
		next if ($line =~ /^\/\//); # TODO: improve C comment filtering

		$line =~ s/\x{FEFF}//g;	# Remove BOM
		$line =~ s/^[\t\s]*//;	# Remove leading tabs and whitespace
		$line =~ s/[\r\n]//g;	# Remove line endings

		my $ptype;

		# struct start
		if($line =~ /^struct (.*) {$/) {
			print "STRUCT: name: " . $1 . "\n";
			$struct->{name} = $1;

		# struct end
		} elsif ($line =~ /}/) {

			if($nestchain && $struct->{packer}{len}) {
				# we get the types from the ones we have read in the packer for the primitive datatypes
				my $types = (ref($struct->{packer}{type}) eq 'ARRAY') ? join(' ', @{$struct->{packer}{type}}) : undef;
				print "hmm\n";			
				my $tempstruct = new Struct(
					handle => $struct->{name},
					types => $types,
					keys => $struct->{packer}{name},		# array of all current primitive datatype names
					len => $struct->{packer}{len},			# FIXME: packets with a* # should be fixed now
					#nested => $struct->{packer}{nested},	# this is never the case imo
				);
				print "yeh\n";
				undef $struct->{packer}{len};
				undef $struct->{packer}{name};
				undef $struct->{packer}{type};

				addNest(\%{$struct->{packer}}, \%{$tempstruct});
			}

			print "STRUCT: end (has $i keys)\n\n";
			my $types = (ref($struct->{packer}{type}) eq 'ARRAY') ? join(' ', @{$struct->{packer}{type}}) : undef;
			$structs->{$struct->{name}} = new Struct(
				handle => $struct->{name},
				types => , $types,
				keys => $struct->{packer}{name},
				len => $struct->{packer}{len},
				nested => $struct->{packer}{nested},
			);
			$i = 0;
			$nestchain = 0;
			$neststruct = 0;
			undef $struct;

		# struct element
		} elsif ($line =~ /\s*([\w\s]*)(\[(.*)\])?;$/) {
			my $text = $1;	# the named vars are just to make it more readable
			my $isArray = $2;
			my $array = $3;

			my @text = split(/ /, $text);
			my $name = pop @text;
			my $ctype = join(" ", @text);
			print "LINE: name: $name, ctype: $ctype\n";

			# type name[x]...;
			if($isArray) {
				my @array = ($array =~ /\]\[/) ? split(/\]\[/, $array) : ($array);
				print "ARRAY: dimension: ".(scalar @array)."\n";
				# char name[x]; //simply a string
				if(scalar @array == 1 && $ctype eq 'char' || $ctype eq 'unsigned char') {	# hack
					$ptype = 'a';
					print (" ° string: " . ($array ? "count: $array" : "slurp") . "\n");
					push @{$struct->{packer}{type}}, ($ptype . ($array ? $array : '*'));
					#if($len_p2b_lut{$ptype}) {
						#$struct->{packer}{len} += ($array ? ($len_p2b_lut{$ptype} * $array) : 0);
						$struct->{packer}{len} = ($array ? ($struct->{packer}{len} + $len_p2b_lut{$ptype} * $array) : undef);
					#} else {
					#	print "OOPS: $ptype\n";
					#}
					push @{$struct->{packer}{name}}, $name;
				# struct structname name[x]...;
				# type mapName[x]...; // all types except char
				} else {
					$neststruct = 1;
					my $isStruct;
					$ptype = type_c2p($ctype, \$isStruct);
					my $j = scalar @array;
					my $hash;
					my $nest;
				
					$hash->{name} = $name;
					$hash->{count} = $array[0];
					if(@array > 1) {
						$hash->{nested} = \%{$nest};
					} elsif ($isStruct) {
						$hash->{nested} = $structs->{$ptype};
					} elsif (@array == 1) {
						$hash->{type} = $ptype;
						$hash->{len} = $len_p2b_lut{$ptype};
					}

					foreach (@array) {
						print (" ° $j-dimensional array: " . ($_ ? "count: $_" : "slurp") . "\n");
						if ($j == scalar @array) {	# first
							$j--;
							next;
						} else {
							$nest->{count} = $_;
							if ($j == 1) {			# last
								if($isStruct) {
									$nest->{nested} = $structs->{$ptype};
								} else {
									$nest->{type} = $ptype;
									$nest->{len} =  $len_p2b_lut{$ptype};
								}
							
							} else {
								$nest = \%{$nest->{nested}};
							}
						}

						$j--;
					}
					# print Dumper($hash);
					if($hash) {
						#print Dumper(arrayMaker(\%{$struct->{packer}{nested}}, \%{$hash})); # shows right
						#print Dumper($struct->{packer}{nested}); # shows wrong, for some reason
						#print "before: " . Dumper($struct->{packer}{nested});
						#print "adding: " . Dumper arrayMaker($struct->{packer}{nested}, \%{$hash});
						addNest($struct->{packer}, arrayMaker($struct->{packer}{nested}, \%{$hash})); # temp hack
						#$struct->{packer}{nested} = arrayMaker(\%{$struct->{packer}{nested}}, \%{$hash}); # temp hack
						#used with the pod'ed sub: arrayMaker(\%{$struct->{packer}}, \%{$hash});
						#print "after: " . Dumper($struct->{packer}{nested}); # shows wrong, for some reason # right now
						$neststruct = 0;
					}
				}
		
			# struct or primitive datatype
			# struct structname name;
			# type name;
			} else {
				my $isStruct;
				$ptype = type_c2p($ctype, \$isStruct);
			
				# structs
				if($isStruct) {
					# we encountered a struct
					# 	make struct from previous and put it in
					# 	put chainmode on for last
					if(ref($struct->{packer}{type}) eq 'ARRAY') {								# now we also do it for packets like PACKET_ZC_ADD_SKILL, in order to not do that we need to be able to look ahead
						print "NEST: nesting $struct->{name} on the same level as $name\n";
						my $types = (ref($struct->{packer}{type}) eq 'ARRAY') ? join(' ', @{$struct->{packer}{type}}) : undef;
						my $tempstruct = new Struct(
							handle => $struct->{name},
							types => $types,
							keys => $struct->{packer}{name},
							len => $struct->{packer}{len},
							#nested => $struct->{packer}{nested}, # never nested imo
						);

						undef $struct->{packer}{len};
						undef $struct->{packer}{name};
						undef $struct->{packer}{type};
						$nestchain = 1;

						addNest(\%{$struct->{packer}}, \%{$tempstruct});
					}
				
					# this is the actual struct we encountered
					#	nest it on the same level as the previous struct we made and nested
					if($name) {
						#if(defined $structs->{'EQUIPSLOTINFO'}) {
						#	print Dumper($structs->{'EQUIPSLOTINFO'});
						#}
						#print "$ptype has defined struct: ".(defined $structs->{$ptype} ? 1 : 0) ."\n";
						#print "handle: " . ((defined $structs->{$ptype}{handle}) ? $structs->{$ptype}{handle} : 0) . "\n";
						#print "name we give: " . $name . "\n";
						#print "types: " . ((defined $structs->{$ptype}{types}) ? $structs->{$ptype}{types} : 0) . "\n";
						#print "keys: " . ((defined $structs->{$ptype}{keys}) ? join(', ', @{$structs->{$ptype}{keys}}) : 0) . "\n";
						#print "len: " . ((defined $structs->{$ptype}{len}) ? $structs->{$ptype}{len} : 0) . "\n";
						my $temp = new Struct(
							handle => (defined $structs->{$ptype}{handle}) ? $structs->{$ptype}{handle} : undef,
							name => $name,
							types => (defined $structs->{$ptype}{types}) ? $structs->{$ptype}{types} : undef,
							keys => (defined $structs->{$ptype}{keys}) ? $structs->{$ptype}{keys} : undef,
							len => (defined $structs->{$ptype}{len}) ? $structs->{$ptype}{len} : undef,
							nested => (defined $structs->{$ptype}{nested}) ? $structs->{$ptype}{nested} : undef,
						);
						addNest(\%{$struct->{packer}}, \%{$temp});
					} else {
						addNest(\%{$struct->{packer}}, \%{$structs->{$ptype}});
					}
					print "NEST: nesting done!\n";
			
				# primitive datatype (not a struct or an array)
				} else {
					push @{$struct->{packer}{type}}, $ptype;
					$struct->{packer}{len} += $len_p2b_lut{$ptype};
					push @{$struct->{packer}{name}}, $name;
				}
			}
		
			$i++;
		}
	}
	$fh->close();		
}

sub addNest {
	my ($nested, $addednest) = @_;
	#print "KEYS: " . join(', ', keys %{$nested->{nested}}) . " " . $nested->{nested}{count} . "\n" if ((ref($nested->{nested})) ne 'ARRAY');
	#print "addNest before: \n" . Dumper ($nested);
	#print "addNest: \n" . Dumper ($addednest);
	debug "addNest: nestchain: $nestchain, neststruct: $neststruct\n";
	if(ref($nested->{nested}) eq 'ARRAY') {
		debug "addNest: is array, push addednest to it.\n";
		push @{$nested->{nested}}, \%{$addednest};

	} elsif (keys %{$nested->{nested}} && ($neststruct || $nestchain)) {
		debug "addNest: exists, make array from nested and add nest.\n";
		$nested->{nested} = [$nested->{nested}, \%{$addednest}];

	} else {
		debug "addNest: does not exist, insert nest.\n";
		$nested->{nested} = \%{$addednest};
		#print "addNest after: \n" . Dumper ($nested);
	}
}

sub arrayMaker { # FIXME: it needs to use the addNest instead of direct access for top, so we can support packet6
	my $nested = shift;
	my $hash = shift;

	if(exists $hash->{nested}) {
		if(blessed($hash->{nested})) {
			debug "arrayMaker: nested blessed: $hash->{count} " .(exists $hash->{name} ? $hash->{name} : ''). "\n";
			$nested = new Array(
				count => $hash->{count},
				name => $hash->{name},
				nested => \%{$hash->{nested}},
			);
		} else {
			debug "arrayMaker: nested: $hash->{count} " .(exists $hash->{name} ? $hash->{name} : ''). "\n";
			$nested = new Array(
				count => $hash->{count},
				name => $hash->{name},
				#nested => arrayMaker(\%{$nested->{nested}}, \%{$hash->{nested}}),
				nested => arrayMaker(\%{$nested->{nested}}, \%{$hash->{nested}}),
			);
		}
	} else {
		if(uc($hash->{type}) eq 'C') {
			debug "arrayMaker: nonest, string: $hash->{count} $hash->{type} $hash->{len}\n";
			$nested = new Scalar(
				count => $hash->{count},
				type => 'a',
				len => $hash->{len},
			);

		} else {
			debug "arrayMaker: nonest, nostring: $hash->{count} $hash->{type} $hash->{len}\n";
			$nested = new Array(
				name => $hash->{name},
				count => $hash->{count},
				type => $hash->{type},
				len => $hash->{len},
			);
		}
	}
	return \%{$nested};
}

sub type_c2p {
	my $ctype = shift;
	my $isStruct = shift;
	my @elements = split(' ', $ctype);
	my $type;

	if (scalar @elements == 1) {
		if ($elements[0] eq 'unsigned') { # I => unsigned (int)
			$type = 'I';
			#$type = $type_c2p_lut{'int'};
		} elsif($type = $type_c2p_lut{$elements[0]}) {
		} else {
			die "INVALID 1!\n";
		}

	} elsif (scalar @elements == 2) {
		if ($elements[0] eq 'signed') {	# signed type
			$type = $type_c2p_lut{$elements[1]};
		} elsif ($elements[0] eq 'unsigned') {	# unsigned type
			$type = uc $type_c2p_lut{$elements[1]};
		} else {
			if ($elements[1] eq 'int') { # type int
				$type = $type_c2p_lut{$elements[0]};
			} elsif ($elements[0] eq 'struct') { # struct
				$type = $elements[1];
				$$isStruct = 1;
			} elsif($type = $type_c2p_lut{$elements[0] . $elements[1]}) { # long long | long double
			} else {
				die "INVALID 2!\n";
			}
		}

	} elsif (scalar @elements == 3) {
		if ($elements[2] ne 'int') { # long long | long double
			$elements[1] .= $elements[2];
		}

		if ($elements[0] eq 'signed') {	# signed type
			$type = $type_c2p_lut{$elements[1]};
		} elsif ($elements[0] eq 'unsigned') {	# unsigned type
			$type = uc $type_c2p_lut{$elements[1]};
		} else {
			die "INVALID 3!\n";
		}
	}

	return $type;
}

#Plugins::callHook('parseMsg/pre', {switch => $switch, msg => $msg, msg_size => length($msg)});
sub parse_packet {
	my $hookName = shift;
	my $args = shift;
    	my $switch = $args->{switch}; # packetID
    	my $unpacked = {};
	my $position = 0;
	return if $switch eq '0087';

	if (exists $packetID2HANDLE->{$args->{switch}}) {
		if (exists $structs->{$packetID2HANDLE->{$args->{switch}}}) {
			message "Writing unpacked struct to DUMP_structs.txt for: $args->{switch}\n";

	    		$structs->{$packetID2HANDLE->{$args->{switch}}}->unpack($unpacked, \$args->{msg}, \$position);

			open DUMP, ">>:utf8", "DUMP_structs.txt";
			print DUMP (uc unpack('H4', pack('n', $unpacked->{PacketType})). "\n" . Dumper($unpacked) . "\n\n\n");
			close DUMP;
		} else {
			error "No unpack object found for: $args->{switch}\n";
		}
	} else {
		error "No packetHANDLER found for: $args->{switch}\n";
	}
}

return 1;
